@@ -10,12 +10,10 @@ import boto3
 import botocore.client
 import six
 
-
 logger = logging.getLogger(__name__)
-logger.addHandler(logging.NullHandler())
 
 # Multiprocessing is unavailable in App Engine (and possibly other sandboxes).
-# The only method currently relying on it is s3_iter_bucket, which is instructed
+# The only method currently relying on it is iter_bucket, which is instructed
 # whether to use it by the MULTIPROCESSING flag.
 _MULTIPROCESSING = False
 try:
@@ -24,10 +22,6 @@ try:
 except ImportError:
     logger.warning("multiprocessing could not be imported and won't be used")
 
-START = 0
-CURRENT = 1
-END = 2
-WHENCE_CHOICES = (START, CURRENT, END)
 
 DEFAULT_MIN_PART_SIZE = 50 * 1024**2
 """Default minimum part size for S3 multipart uploads"""
@@ -39,12 +33,22 @@ MODES = (READ_BINARY, WRITE_BINARY)
 """Allowed I/O modes for working with S3."""
 
 BINARY_NEWLINE = b'\n'
-DEFAULT_BUFFER_SIZE = 128 * 1024
 
 SUPPORTED_SCHEMES = ("s3", "s3n", 's3u', "s3a")
 
+DEFAULT_BUFFER_SIZE = 128 * 1024
 
-def _range_string(start, stop=None):
+START = 0
+CURRENT = 1
+END = 2
+WHENCE_CHOICES = [START, CURRENT, END]
+
+
+def clamp(value, minval, maxval):
+    return max(min(value, maxval), minval)
+
+
+def make_range_string(start, stop=None):
     #
     # https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.35
     #
@@ -53,25 +57,60 @@ def _range_string(start, stop=None):
     return 'bytes=%d-%d' % (start, stop)
 
 
-def _clamp(value, minval, maxval):
-    return max(min(value, maxval), minval)
+def open(
+        bucket_id,
+        key_id,
+        mode,
+        buffer_size=DEFAULT_BUFFER_SIZE,
+        min_part_size=DEFAULT_MIN_PART_SIZE,
+        session=None,
+        resource_kwargs=dict(),
+        multipart_upload_kwargs=dict(),
+        ):
+    """Open an S3 object for reading or writing.
 
+    Parameters
+    ----------
+    bucket_id: str
+        The name of the bucket this object resides in.
+    key_id: str
+        The name of the key within the bucket.
+    mode: str
+        The mode for opening the object.  Must be either "rb" or "wb".
+    buffer_size: int, optional
+        The buffer size to use when performing I/O.
+    min_part_size: int, optional
+        The minimum part size for multipart uploads.  For writing only.
+    session: object, optional
+        The S3 session to use when working with boto3.
+    resource_kwargs: dict, optional
+        Keyword arguments to use when creating a new resource.  For writing only.
+    multipart_upload_kwargs: dict, optional
+        Additional parameters to pass to boto3's initiate_multipart_upload function.
+        For writing only.
 
-def open(bucket_id, key_id, mode, **kwargs):
+    """
     logger.debug('%r', locals())
     if mode not in MODES:
         raise NotImplementedError('bad mode: %r expected one of %r' % (mode, MODES))
 
-    encoding = kwargs.pop("encoding", "utf-8")
-    errors = kwargs.pop("errors", None)
-    newline = kwargs.pop("newline", None)
-    line_buffering = kwargs.pop("line_buffering", False)
-    s3_min_part_size = kwargs.pop("s3_min_part_size", DEFAULT_MIN_PART_SIZE)
-
     if mode == READ_BINARY:
-        fileobj = SeekableBufferedInputBase(bucket_id, key_id, **kwargs)
+        fileobj = SeekableBufferedInputBase(
+            bucket_id,
+            key_id,
+            buffer_size=buffer_size,
+            session=session,
+            resource_kwargs=resource_kwargs,
+        )
     elif mode == WRITE_BINARY:
-        fileobj = BufferedOutputBase(bucket_id, key_id, min_part_size=s3_min_part_size, **kwargs)
+        fileobj = BufferedOutputBase(
+            bucket_id,
+            key_id,
+            min_part_size=min_part_size,
+            session=session,
+            multipart_upload_kwargs=multipart_upload_kwargs,
+            resource_kwargs=resource_kwargs,
+        )
     else:
         assert False, 'unexpected mode: %r' % mode
 
@@ -108,7 +147,7 @@ class SeekableRawReader(object):
         :param int position: The byte offset from the beginning of the key.
         """
         self._position = position
-        range_string = _range_string(self._position)
+        range_string = make_range_string(self._position)
         logger.debug('content_length: %r range_string: %r', self._content_length, range_string)
 
         #
@@ -142,12 +181,10 @@ class SeekableRawReader(object):
 
 class BufferedInputBase(io.BufferedIOBase):
     def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
-                 line_terminator=BINARY_NEWLINE, **kwargs):
-        session = kwargs.pop(
-            's3_session',
-            boto3.Session(profile_name=kwargs.pop('profile_name', None))
-        )
-        s3 = session.resource('s3', **kwargs)
+                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=dict()):
+        if session is None:
+            session = boto3.Session()
+        s3 = session.resource('s3', **resource_kwargs)
         self._object = s3.Object(bucket, key)
         self._raw_reader = RawReader(self._object)
         self._content_length = self._object.content_length
@@ -283,12 +320,10 @@ class SeekableBufferedInputBase(BufferedInputBase):
     Implements the io.BufferedIOBase interface of the standard library."""
 
     def __init__(self, bucket, key, buffer_size=DEFAULT_BUFFER_SIZE,
-                 line_terminator=BINARY_NEWLINE, **kwargs):
-        session = kwargs.pop(
-            's3_session',
-            boto3.Session(profile_name=kwargs.pop('profile_name', None))
-        )
-        s3 = session.resource('s3', **kwargs)
+                 line_terminator=BINARY_NEWLINE, session=None, resource_kwargs=dict()):
+        if session is None:
+            session = boto3.Session()
+        s3 = session.resource('s3', **resource_kwargs)
         self._object = s3.Object(bucket, key)
         self._raw_reader = SeekableRawReader(self._object)
         self._content_length = self._object.content_length
@@ -326,7 +361,7 @@ class SeekableBufferedInputBase(BufferedInputBase):
             new_position = self._current_pos + offset
         else:
             new_position = self._content_length + offset
-        new_position = _clamp(new_position, 0, self._content_length)
+        new_position = clamp(new_position, 0, self._content_length)
         self._current_pos = new_position
         self._raw_reader.seek(new_position)
         logger.debug('new_position: %r', self._current_pos)
@@ -349,16 +384,24 @@ class BufferedOutputBase(io.BufferedIOBase):
 
     Implements the io.BufferedIOBase interface of the standard library."""
 
-    def __init__(self, bucket, key, min_part_size=DEFAULT_MIN_PART_SIZE, s3_upload=None, **kwargs):
+    def __init__(
+            self,
+            bucket,
+            key,
+            min_part_size=DEFAULT_MIN_PART_SIZE,
+            s3_upload=None,
+            session=None,
+            resource_kwargs=dict(),
+            multipart_upload_kwargs=dict(),
+            ):
         if min_part_size < MIN_MIN_PART_SIZE:
             logger.warning("S3 requires minimum part size >= 5MB; \
 multipart upload may fail")
 
-        session = kwargs.pop(
-            's3_session',
-            boto3.Session(profile_name=kwargs.pop('profile_name', None))
-        )
-        s3 = session.resource('s3', **kwargs)
+        if session is None:
+            session = boto3.Session()
+
+        s3 = session.resource('s3', **resource_kwargs)
 
         #
         # https://stackoverflow.com/questions/26871884/how-can-i-easily-determine-if-a-boto-3-s3-bucket-resource-exists
@@ -369,7 +412,7 @@ multipart upload may fail")
             raise ValueError('the bucket %r does not exist, or is forbidden for access' % bucket)
         self._object = s3.Object(bucket, key)
         self._min_part_size = min_part_size
-        self._mp = self._object.initiate_multipart_upload(**(s3_upload or {}))
+        self._mp = self._object.initiate_multipart_upload(**multipart_upload_kwargs)
 
         self._buf = io.BytesIO()
         self._total_bytes = 0
@@ -477,23 +520,43 @@ multipart upload may fail")
             self.close()
 
 
-def iter_bucket(bucket_name, prefix='', accept_key=lambda key: True,
+def iter_bucket(bucket_name, prefix='', accept_key=None,
                 key_limit=None, workers=16, retries=3):
     """
-    Iterate and download all S3 files under `bucket/prefix`, yielding out
-    `(key, key content)` 2-tuples (generator).
+    Iterate and download all S3 objects under `s3://bucket_name/prefix`.
 
-    `accept_key` is a function that accepts a key name (unicode string) and
-    returns True/False, signalling whether the given key should be downloaded out or
-    not (default: accept all keys).
+    Parameters
+    ----------
+    bucket_name: str
+        The name of the bucket.
+    prefix: str, optional
+        Limits the iteration to keys starting wit the prefix.
+    accept_key: callable, optional
+        This is a function that accepts a key name (unicode string) and
+        returns True/False, signalling whether the given key should be downloaded.
+        The default behavior is to accept all keys.
+    key_limit: int, optional
+        If specified, the iterator will stop after yielding this many results.
+    workers: int, optional
+        The number of subprocesses to use.
+    retries: int, optional
+        The number of time to retry a failed download.
 
-    If `key_limit` is given, stop after yielding out that many results.
+    Yields
+    ------
+    str
+        The full key name (does not include the bucket name).
+    bytes
+        The full contents of the key.
 
+    Notes
+    -----
     The keys are processed in parallel, using `workers` processes (default: 16),
     to speed up downloads greatly. If multiprocessing is not available, thus
     _MULTIPROCESSING is False, this parameter will be ignored.
 
-    Example::
+    Examples
+    --------
 
       >>> # get all JSON files under "mybucket/foo/"
       >>> for key, content in iter_bucket(bucket_name, prefix='foo/', accept_key=lambda key: key.endswith('.json')):
@@ -503,6 +566,9 @@ def iter_bucket(bucket_name, prefix='', accept_key=lambda key: True,
       >>> for key, content in iter_bucket(bucket_name, key_limit=10000, workers=32):
       ...     print key, len(content)
     """
+    if accept_key is None:
+        accept_key = lambda key: True
+
     #
     # If people insist on giving us bucket instances, silently extract the name
     # before moving on.  Works for boto3 as well as boto.
